
/****************************************************************************
 *                                                                          *
 *     Numerical Methods subroutines                                        *
 *                                                                          *
 ****************************************************************************/
// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "Constants.h"
#include "Externals.h"
#include "nrutil.h"

#define NRANSI

/* Newton-Raphson method, called by solve.C
 * Return: 0=OK !0=ERROR
 */
int mnewt(int ntrial, int n, double tolx, double tolf, 
	  int *iterations, double *x)
{
  void lubksb(double **a, int n, int *indx, double b[]);
  void ludcmp(double **a, int n, int *indx, double *d);
  extern void usrfun(int n, double *fvec, double **fjac, double *x, 
		     int iterations);
  int    k, i, *indx;
  double errx, errf, d, *fvec, **fjac, *p;
  
  indx=ivector(1,n+1);
  p   =dvector(1,n+1);
  fvec=dvector(1,n+1);
  fjac=dmatrix(1,n+1,1,n+1); // jacobi matrix

  for (k=1;k<=ntrial;k++) {
    if (Control.output > 3 && k<=10) {
      double xl=0.0;
      printf("mnewt[%d].x:\n",k);
      for(i=1;i<=n;i++) {
        xl+=x[i]*x[i]; /* norm */
        if (i==6 && 10<=n) printf(" ...");
        if (i<=5 || i>n-6) printf(" % 10.8f",x[i]);
      }
      printf("\nmnewt[%d].x^2: % 10.8e\n",k,xl);
    }
    *iterations = k;
    usrfun(n,fvec,fjac,x,*iterations);  // out: fvec[n], fjac[n][n],  in: terms*x[n]
    errf=0.0;
    for (i=1;i<=n;i++) errf += fabs(fvec[i]);
    if (Control.output > 1)                                                       // JS1002 2 to 1
      printf(" %4d mnewt.errf= %8.2e %8.2e x[1]=%14.12f\n", k, errf, tolf, x[1]); // JS1002 %16.13 to %7.1
    if (errf <= tolf) break;  // convergence reached (normally we leave the for-loop here)
    for (i=1;i<=n;i++) p[i] = -fvec[i];
    ludcmp(fjac,n,indx,&d); // in: fjac[][]             out: fjac[][],indx[],d (sort columns)
    lubksb(fjac,n,indx,p);  // in: fjac[][],indx[],p[]  out: p[] (SolveEq: Ax=b p.o=x p.i=b fjac=A)
    errx=0.0;
    for (i=1;i<=n;i++) {
      errx += fabs(p[i]);
      x[i] += p[i];
    }
    if (Control.output > 1)
      printf(" %4d mnewt.errx= %8.2e %8.2e x[1]=%14.12f\n", k, errx, tolx, x[1]); // JS1002 %16.13 to %7.1
    if (errx <= tolx) break;
    if (errx > 1e+10) { *iterations=(k+=ntrial); } // abort, no convergence
  }
  if (k>=ntrial) for (i=1;i<=n;i++) x[i]=0.0;  // Oct09 JS define as bad
  free_dmatrix(fjac,1,n+1,1,n+1);
  free_dvector(fvec,1,n+1);
  free_dvector(p,1,n+1);
  free_ivector(indx,1,n+1);
  if (*iterations>=ntrial) return -1; // JS091021: return error state
  return 0;
}
#undef NRANSI

// LUBKSB Procedure from Numerical Recipes in C
//    Solves the set of n linear equations Ax = b.
//    (LUBKSB must be used with the procedure LUDCMP to do this.)
// in:  a The LU decomposition of a matrix, created by LUDCMP. Parameter a
//        is not modified by calling this procedure.
//        index A vector, created by LUDCMP, containing the row
//        permutations effected by the partial pivoting.
//      b On input, b contains the vector on the right-hand side of the
//        equation. Must be of data type FLOAT (changed to double!)
// out: b On output, b is replaced by the solution, vector x.
//      LUBKSB must be used in conjunction with LUDCMP.
// in: a[1..n][1..n], b[1..n], indx[1..n],  out: b[1..n]
void lubksb(double **a, int n, int *indx, double b[])
{
  int i,ii=0,ip,j;
  double sum;
  
  for (i=1;i<=n;i++) {
    ip=indx[i];
    sum=b[ip];
    b[ip]=b[i];
    if (ii)
      for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
    else if (sum) ii=i;
    b[i]=sum;
  }
  for (i=n;i>=1;i--) {
    sum=b[i];
    for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
    b[i]=sum/a[i][i];
  }
}

#define NRANSI
#define TINY 1.0e-20;

// input:  n = matrix size, a[1..n][1..n] 
// output: a[1..n][1..n], d={-1,+1}, indx[1..n]
// ToDo: check taken from NumRec? JS
void ludcmp(double **a, int n, int *indx, double *d)
{
  int i,imax,j,k;
  double big,dum,sum,temp;
  static double *vv=NULL;
  static int oldn=0;
  
  if (vv==NULL || oldn!=n) {
    if (vv) free_dvector(vv,1,oldn+1);
    vv=dvector(1,n+1); oldn=n;  // call it only if n has changed, Joerg Sep05
  }
  *d=1.0;
  for (i=1;i<=n;i++) {
    for (big=0.0,j=1;j<=n;j++) // maximum element of line i = big
      if ((temp=fabs(a[i][j])) > big) big=temp;
    if (big == 0.0) {
      printf("Singular matrix in routine ludcmp, a[%d][1..%d]==0\n", i, n);
      // exit(1); // was nrerror (ToDo: better way?)
      vv[i]=0;  // works for kagome_triangle.C D=0 Mag[1..3]!=Mag[4]=free
                // not for triangle_ext_field_2 bad startvec + LSUB4
      continue;
    }
    vv[i]=1.0/big;  /* normalize? */
  }
  for (j=1;j<=n;j++) {
    for (i=1;i<j;i++) {
      sum=a[i][j];
      for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
    }
    big=0.0;
    for (imax=i=j;i<=n;i++) { // JS060711 imax= added
      sum=a[i][j];
      for (k=1;k<j;k++) sum -= a[i][k]*a[k][j];
      a[i][j]=sum;
      if ( (dum=vv[i]*fabs(sum)) >= big) {
	big=dum;
	imax=i;
      }
    }
    if (j != imax) {  // exchange column imax with j
      for (k=1;k<=n;k++) {
	dum=a[imax][k]; a[imax][k]=a[j][k]; a[j][k]=dum;
      }
      *d = -(*d);
      vv[imax]=vv[j];
    }
    indx[j]=imax;
    if (a[j][j] == 0.0) a[j][j]=TINY;
    if (j != n) {
      dum=1.0/(a[j][j]);
      for (i=j+1;i<=n;i++) a[i][j] *= dum;
    }
  }
  // free_dvector(vv,1,n+1);
}
#undef TINY
#undef NRANSI

